from multiprocessing import Process
import os
import sys
import shlex
from ctypes import cdll
import time
import signal
import select
import re
import traceback
import statistics

# /usr/include/linux/prctl.h
PR_SET_PDEATHSIG = 1

class Workflow:
    def __init__(self, hosts_data, timeout):
        self.debug = False
        self.hosts_data = hosts_data
        self.timeout = timeout

    def child_process(self, cmdline, ppid):
        # no effect for setuid or binaries with capabilities!
        errno = cdll['libc.so.6'].prctl(PR_SET_PDEATHSIG, signal.SIGTERM)
        if errno != 0:
            raise Exception(f'prctl() failed with errno={errno}')
        if os.getppid() != ppid:
            raise Exception('Parent already terminated')

        os.close(0) # stdin

        args = shlex.split(cmdline)
        os.execvp(args[0], args)
        # never returns

    def start_process(self, cmdline):
        (pipe_r, pipe_w) = os.pipe()
        parent_pid = os.getpid()

        # Python runs only one thread at a time
        # so we should be safe to execute fork().
        # Once a process fork()'ed, by definition
        # parent threads are not inherited.
        pid = os.fork()
        if pid == 0:
            # Make sure that we don't print anything to stdout/err from this uncontrolled child
            # because it confuses the fullscreen terminal.
            try:
                # Therefore, ensure that we first redirect stdout/err to the pipe.
                os.dup2(pipe_w, 1) # stdout
                os.dup2(pipe_w, 2) # stderr
                os.close(pipe_r)
            except:
                os._exit(254)
            # stdout/err are redirected; we can safely print() anything now
            try:
                # Only the following function must be in this try..except block
                # because we always peek 2 tracebacks back upon exception.
                self.child_process(cmdline, parent_pid)
            except:
                ex_type, ex_value, ex_traceback = sys.exc_info()
                (tb_file, tb_line, tb_func, tb_text) = traceback.extract_tb(ex_traceback)[2]
                print(f'Fatal error: {tb_func}(): {ex_type.__name__}: {ex_value}')
                os._exit(255)
        else:
            os.close(pipe_w)

        return pid, pipe_r

    def start_all_processes(self):
        self.fd_lookup = {}

        for hostname in self.hosts_data:
            data = self.hosts_data[hostname]
            pid, pipe_r = self.start_process(data['proc']['cmdline'])
            data['proc']['pid'] = pid
            data['proc']['out_fd'] = pipe_r
            self.fd_lookup[pipe_r] = hostname

        self.exited_hosts = []
        self.select_fdlist = list(self.fd_lookup.keys())

    def parse_time(self, line):
        line = line.strip()
        if not len(line):
            return None

        if re.search(r'^PING\s.+((bytes of data)|(data bytes))', line):
            return None

        m = re.search(r'^\d+\sbytes\sfrom\s.+\sttl=\d+\s+time=([\d\.]+)\sms$', line)
        if not m:
            res = '???'
        else:
            res = m.group(1)
            try:
                res = float(res)
                res = round(res)
            except:
                res = 'ERR' # this should never happen

        return res

    def parse_seq(self, line):
        m = re.search(r'\sicmp_seq=(\d+)(\s|$)', line)
        if m:
            return int(m.group(1))
        else:
            return None

    def parse_timeout(self, line):
        return re.search(r'^no answer yet for icmp_seq=\d+$', line)

    def handle_pipes(self, timeout):
        if not len(self.select_fdlist):
            time.sleep(0.05)
            if self.debug:
                print('== No active processes')
                time.sleep(1)
            return

        (fds_read_ready, _, _) = select.select(self.select_fdlist, [], [], timeout)
        for fd in fds_read_ready:
            s = os.read(fd, 1024 * 1024)

            if not len(s): # EOF
                terminated = True
                s = '\nCommand terminated.\n'
                self.select_fdlist.remove(fd)
                self.exited_hosts.append(self.fd_lookup[fd])
            else:
                terminated = False
                s = s.decode('ascii', 'replace')

            all_s_parts = s.split('\n')

            s_ends_newline = s.endswith('\n')
            if s_ends_newline:
                all_s_parts.pop() # remove this empty line which shows that "s" ends in "\n"

            data = self.hosts_data[self.fd_lookup[fd]]
            with data['lock']:
                for idx, part in enumerate(all_s_parts): # all we have left here is real data
                    if idx != len(all_s_parts) - 1: # not last element
                        newline = True # all non-last elements ended in "\n" and were split
                    else: # last element
                        newline = s_ends_newline

                    if not data['raw_complete']: # last line didn't end with "\n"
                        data['raw'][-1] += part # append to the existing line
                    else: # last line was ended with "\n"
                        data['raw'].append(part) # start a new line

                    data['raw_complete'] = newline

                    if newline:
                        if self.debug:
                            print(data['raw'][-1])

                        if not terminated:
                            seq = self.parse_seq(data['raw'][-1])

                            if seq is not None:
                                if data['seen_rx_seq'].get(seq):
                                    pd = self.parse_time(data['raw'][-1])
                                    if pd:
                                        # display the raw "time" value in the "Last" stats
                                        # even if it was marked as a timeout already
                                        data['stats']['Last'] = pd
                                    continue # we have already handled this "seq"
                                else:
                                    data['seen_rx_seq'][seq] = True

                            if seq is not None and seq > data['stats']['TX_cnt']:
                                data['stats']['TX_cnt'] = seq

                            is_timeout = self.parse_timeout(data['raw'][-1])

                            if is_timeout:
                                pd = '*'
                            else:
                                pd = self.parse_time(data['raw'][-1])

                            data['stats']['Last'] = pd
                            if pd is not None:
                                data['parsed'].append(pd)
                                if self.debug:
                                    print(f'PARSED: "{pd}"')

                            if type(pd) is int and pd < self.timeout * 1000:
                                data['stats']['RX_cnt'] += 1
                            data['stats']['XX_cnt'] = \
                                data['stats']['TX_cnt'] - data['stats']['RX_cnt']

                            if data['stats']['TX_cnt'] > 0:
                                data['stats']['Loss%'] = '{:.0f}%'.format(
                                    (1 - data['stats']['RX_cnt'] / data['stats']['TX_cnt']) * 100
                                )

                            points = list(filter(lambda x: type(x) is int, data['parsed']))
                            if len(points):
                                data['stats']['StDev'] = '{:.1f}'.format(statistics.pstdev(points))
                                data['stats']['Max'] = max(points)
                                data['stats']['Min'] = min(points)
                                data['stats']['Avg'] = round(sum(points) / len(points))

    def handle_exited_hosts(self):
        done_hostnames = []
        for hostname in self.exited_hosts:
            data = self.hosts_data[hostname]
            
            einfo = os.waitid(os.P_PID, data['proc']['pid'], os.WEXITED | os.WNOHANG)
            if einfo is None: # child not completely exited, yet
                continue

            done_hostnames.append(hostname)

            if einfo.si_code == os.CLD_EXITED:
                term_reason = f'exited with status {einfo.si_status}'
            else:
                term_reason = f'killed by signal {einfo.si_status}' # (si_code={einfo.si_code})

            with data['lock']:
                data['proc']['pid'] = None
                data['raw'].append(f'== Process {term_reason}')
                data['parsed'].append('EXIT')
                if self.debug:
                    print(data['raw'][-1])

        for hostname in done_hostnames:
            self.exited_hosts.remove(hostname)

    def update_hosts_data(self, timeout):
        self.handle_pipes(timeout)
        self.handle_exited_hosts()



import argparse
import shlex
import ping_multi_ext # version

def argv_parser_base(prog_desc):
    parser = argparse.ArgumentParser(
        description=prog_desc
    )
    vstr = '{} {} | {}'.format(
        '%(prog)s', ping_multi_ext.version,
        'https://github.com/famzah/ping-multi-ext'
    )
    parser.add_argument('--version', action='version', version=vstr)

    return parser

def remove_ssh_user(host):
    parts = host.split('@', 2)
    
    if len(parts) == 3:
        parts.pop(1)

    return '@'.join(parts)

def compose_ping_cmd(host, cmd_args):
    parts = host.split('@', 1)

    ping_cmd = 'ping -O -W {} -i {} {}'.format(
        shlex.quote(str(cmd_args['wait'])),
        shlex.quote(str(cmd_args['interval'])),
        shlex.quote(parts[0])
    )

    if len(parts) > 1:
        return 'ssh -o BatchMode=yes {} {}'.format(
            shlex.quote(parts[1]),
            shlex.quote(ping_cmd)
        )
    else:
        return ping_cmd



import ping_multi_ext.lib
import ping_multi_ext.core

def parse_argv():
    parser = ping_multi_ext.lib.argv_parser_base(
        'Execute multiple external ping commands at once.'
    )

    # non-required first

    dval = 1
    parser.add_argument('--timeout', type=float, default=dval,
        help=f'ping reply timeout in seconds; default={dval}')

    dval = 0
    parser.add_argument('--hosts-max-width', type=int, default=dval,
        help=f'maximum width of the hosts column; default={dval}')

    # required

    parser.add_argument('--ping', nargs=2, metavar=('UNIQUE_NAME', 'COMMAND'),
        action='append', required=True,
        help='each command must be specified by an arbitrary unique name and the command itself; ' +\
             'you can specify this option many times'
    )

    return vars(parser.parse_args())

def main():
    ping_multi_ext.core.main(parse_argv())



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.rst').read_text(encoding='utf-8')

import ping_multi_ext # version

setup(
    name='ping-multi-ext',
    version=ping_multi_ext.version,

    description='Interactively ping one host from multiple locations via SSH',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/famzah/ping-multi-ext',

    author='Ivan Zahariev (famzah)',
    #author_email='author@example.com',  # Optional

    # list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Internet',
        'Topic :: System :: Networking :: Monitoring',
        'Topic :: Terminals',
        'Topic :: Utilities',
    ],

    keywords='ping, multi, console, ssh, terminal, interactive',

    # When your source code is in a subdirectory under the project root, e.g.
    # `src/`, it is necessary to specify the `package_dir` argument.
    #package_dir={'': 'src'},  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    #packages=find_packages(where='src'),  # Required
    packages=['ping_multi_ext'],

    python_requires='>=3.6, <4',
    install_requires=['blessings', 'curtsies'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    package_data={
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/distutils/setupscript.html#installing-additional-files
    #
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[],

    entry_points={
        'console_scripts': [
            'ping-raw-multi=ping_multi_ext.cmd_raw:main',
            'ping-multi=ping_multi_ext.cmd_multi:main',
        ],
    },

    project_urls={
        'Bug Reports': 'https://github.com/famzah/ping-multi-ext/issues',
        'Source': 'https://github.com/famzah/ping-multi-ext',
    },
)



from blessings import Terminal
import time
import threading
import sys
import curtsies
import signal
import random
import string
import math
import ping_multi_ext.proc
from collections import deque
import os

gvars = {}

class TermCtrl:
    def __init__(self, key):
        self.code = getattr(gvars['term'], key)

def _compose_printed_str(data_row, t_width):
    printed_str = ''
    pstr_len = 0

    for part_idx, part_value in enumerate(data_row):
        if isinstance(part_value, TermCtrl):
            printed_str += part_value.code
        elif isinstance(part_value, str):
            if pstr_len + len(part_value) <= t_width:
                printed_str += part_value
                pstr_len += len(part_value)
            else: # we are going to exceed the width
                cut_value = part_value[0:(t_width - pstr_len)]

                printed_str += cut_value
                pstr_len += len(cut_value)

                break # stop processing the input data
        else:
            raise NotImplementedError(part_value)

    if pstr_len > t_width:
        raise Exception(f'Assertion failed: "{printed_str}"')

    return printed_str

def _sanity_check_data_row(data_row, i):
    for part_idx, part_value in enumerate(data_row):
        part_str_id = f'Screen data row with index {i} part {part_idx}'

        if not isinstance(part_value, str) and not isinstance(part_value, TermCtrl):
            raise ValueError(f'{part_str_id} is not "str" nor "TermCtrl"')
        if isinstance(part_value, str):
            for c in ['\n', '\r', '\t']:
                if c in part_value:
                    raise ValueError(
                        f'{part_str_id} cannot contain the ASCII character {ord(c)}: "{part_value}"'
                    )

def ui_print(term, data):
    # get it every time, so that we react if the terminal changes (ie. if it gets resized)
    t_height = term.height
    t_width = term.width

    term_size_changed = False
    if 'ui_old_size' not in gvars \
      or t_height != gvars['ui_old_size']['height'] or t_width != gvars['ui_old_size']['width']:
        #
        term_size_changed = True
        gvars['ui_old_size'] = {
            'height': t_height,
            'width': t_width,
        }

    if term_size_changed:
        print(term.clear(), end='', flush=True)
        if 'ui_old_screen' in gvars:
            del gvars['ui_old_screen']

    if 'ui_old_screen' not in gvars:
        gvars['ui_old_screen'] = {}
    old_screen = gvars['ui_old_screen']

    screen_changed = False
    for i in range(0, t_height):
        if i not in old_screen: # first initialization of "old_screen"
            old_screen[i] = None

        try:
            data_row = data[i]
        except IndexError:
            data_row = ['']

        _sanity_check_data_row(data_row, i)

        printed_str = _compose_printed_str(data_row, t_width)

        if printed_str != old_screen[i]: # this could be optimized to draw only a part of the line
            old_screen[i] = printed_str
            screen_changed = True
            # always restore the color to "normal" in case we cut this in _compose_printed_str()
            # always clear the line until the end in case we're printing something shorter now
            print(term.move(i, 0) + printed_str + term.normal + term.clear_eol, end='', flush=False)

    if screen_changed:
        sys.stdout.flush()

    return term_size_changed, t_height, t_width

def _compose_host_data_parsed_str(hostname, host_data, t_width, selected):
    row_parts = []
    row_parts_str_len = 0

    if selected:
        row_parts.append(TermCtrl('bold'))

    host_id_str = ('{:<' + str(gvars['config']['max_host_id_len']) + 's} ').format(
        hostname[0:gvars['config']['max_host_id_len']]
    )
    row_parts.append(host_id_str)
    row_parts_str_len += len(host_id_str)

    s = ''
    with host_data['lock']:
        stats_val = host_data['stats'][gvars['stats_show'][0]]
        if stats_val is None:
            stats_val = ''

        stats_str = ('{:>6}  '.format(stats_val))
        row_parts.append(stats_str)
        row_parts_str_len += len(stats_str)

        for v_idx in range(1, len(host_data['parsed']) + 1):
            raw_value = host_data['parsed'][-v_idx]

            if gvars['time_scale'][0] == 'success' or gvars['time_scale'][0] == 'numbered':
                try:
                    int_v = int(raw_value)
                except:
                    added_value = raw_value.strip()[0:1]
                    if added_value == '*': # timeout
                        if gvars['time_scale'][0] == 'success':
                            added_value = 'X' # use the same symbol
                        if gvars['time_scale'][0] == 'numbered':
                            added_value = '-' # use the same symbol
                else:
                    if gvars['time_scale'][0] == 'success':
                        if int_v < gvars['cmd_args']['timeout'] * 1000:
                            added_value = '.'
                        else:
                            added_value = 'X'
                    elif gvars['time_scale'][0] == 'numbered':
                        if int_v < 1000 and int_v < gvars['cmd_args']['timeout'] * 1000:
                            added_value = math.floor(int_v / 100)
                        else:
                            added_value = '-'
                    else:
                        raise ValueError(gvars['time_scale'][0])
            elif gvars['time_scale'][0] == 'raw':
                if type(raw_value) is int and raw_value >= gvars['cmd_args']['timeout'] * 1000:
                    added_value = '*'
                else:
                    added_value = raw_value

                added_value = '{:>4} '.format(added_value)
            else:
                raise ValueError(gvars['time_scale'][0])

            added_value = str(added_value)

            if len(s) + len(added_value) > t_width - row_parts_str_len:
                break

            s += added_value
    row_parts.append(s)

    return row_parts

def _ui_render_all_hosts_data(
        screen_rows, all_hosts, host_data_type,
        min_idx, max_idx, sel_idx, sel_hostname, t_width
    ):
        if host_data_type == 'parsed':
            for idx, hostname in enumerate(list(gvars['hosts_print_order'])):
                if idx < min_idx:
                    continue
                if idx > max_idx:
                    continue

                screen_rows.append(_compose_host_data_parsed_str(
                    hostname, all_hosts[hostname],
                    t_width, sel_idx == idx
                ))
        elif host_data_type == 'raw':
            for idx, data_row in enumerate(list(all_hosts[sel_hostname]['raw'])):
                if idx < min_idx:
                    continue
                if idx > max_idx:
                    continue
                
                safe_s = str(data_row)
                for c in ['\n', '\r', '\t']:
                    safe_s = safe_s.replace(c, ' ')
                screen_rows.append(safe_s)
        else:
            raise NotImplementedError(host_data_type)

def _ui_render_header(screen_rows, cmd_err, host_data_type, sel_hostname):
    first_row = [
        'Keys: ',
    ]

    if host_data_type == 'parsed':
        first_row.extend([
            TermCtrl('bold'), '↑', TermCtrl('normal'),
            '/',
            TermCtrl('bold'), '↓', TermCtrl('normal'),
            '/',
            TermCtrl('bold'), 'Enter', TermCtrl('normal'),
            '/',
        ])

    first_row.extend([
        TermCtrl('bold'), 'PgUp', TermCtrl('normal'),
        '/',
        TermCtrl('bold'), 'PgDn', TermCtrl('normal'),
    ])

    if host_data_type == 'raw':
        first_row.extend([
            '/',
            TermCtrl('bold'), 'Home', TermCtrl('normal'),
            '/',
            TermCtrl('bold'), 'End', TermCtrl('normal'),
            '/',
            TermCtrl('bold'), 'ESC', TermCtrl('normal'),
        ])

    if host_data_type == 'parsed':
        first_row.extend([
            ' | ',
            TermCtrl('bold'), 'S', TermCtrl('normal'),
            'tats | ',
            TermCtrl('bold'), 'T', TermCtrl('normal'),
            'ime scale',
        ])

    first_row.extend([
        ' | ',
        TermCtrl('bold'), 'Q', TermCtrl('normal'), 'uit',
    ])

    screen_rows.append(first_row)

    if host_data_type == 'parsed':
        if gvars['time_scale'][0] == 'success':
            res_header = ['Ping results', TermCtrl('normal'), ' (success, newest first)']
        elif gvars['time_scale'][0] == 'numbered':
            res_header = ['Ping results', TermCtrl('normal'), ' (scaled per 100 ms, newest first)']
        elif gvars['time_scale'][0] == 'raw':
            res_header = ['Ping results', TermCtrl('normal'), ' (RTT, newest first)']
        else:
            raise ValueError(gvars['time_scale'][0])
        screen_rows.append(
            [
                TermCtrl('bold'),
                ('{:<' + str(gvars['config']['max_host_id_len']) + 's} ').format('Hostname'),
                ('{:>6s}  ').format(gvars['stats_show'][0])
            ] +\
            res_header
        )
    elif host_data_type == 'raw':
        screen_rows.append([TermCtrl('bold'), f'Raw ping results for "{sel_hostname}"'])
    else:
        raise NotImplementedError(host_data_type)

    # insert as second row (moving any other existing headers next)
    screen_rows.insert(1, cmd_err)

class DataScroller:
    def __init__(self, hosts_print_order, tail_mode_cap):
        self.hosts_print_order = hosts_print_order
        self.initialized = False
        self.tail_mode_cap = tail_mode_cap
    
    def set_data_items_count(self, n):
        self.data_items_count = n

    def set_avail_term_rows(self, n):
        self.avail_term_rows = n
        self.scroll_by = max(1, int(n / 2.0))

    def update_max_id(self):
        self.max_idx = min(
            self.min_idx + self.avail_term_rows - 1,
            self.data_items_count - 1
        )

    def reset(self):
        self.initialized = True
        self.in_tail_mode = self.tail_mode_cap

        self.min_idx = 0
        self.sel_idx = None
        self.update_max_id()

    def key_down(self):
        self.in_tail_mode = False
        if self.sel_idx is not None:
            self.sel_idx += 1
        if self.sel_idx is None or self.sel_idx > self.max_idx:
            self.sel_idx = self.min_idx

    def key_up(self):
        self.in_tail_mode = False
        if self.sel_idx is not None:
            self.sel_idx -= 1
        if self.sel_idx is None or self.sel_idx < self.min_idx:
            self.sel_idx = self.max_idx

    def key_pageup(self):
        self.in_tail_mode = False
        if self.min_idx > 0:
            self.min_idx -= self.scroll_by
            self.min_idx = self._align_value_to(self.min_idx, self.scroll_by)

            self.update_max_id()
            self.sel_idx = None

    def key_pagedown(self):
        self.in_tail_mode = False
        if self.max_idx < self.data_items_count - 1:
            self.min_idx += self.scroll_by
            self.update_max_id()
            self.sel_idx = None

        if self.max_idx == self.data_items_count - 1: # we're already at the very end
            if self.tail_mode_cap:
                self.in_tail_mode = True

    def key_home(self):
        self.reset()
        self.in_tail_mode = False

    def key_end(self):
        if self.tail_mode_cap:
            self.in_tail_mode = True

        self.max_idx = self.data_items_count - 1

        self.min_idx = max(0, self.max_idx - self.avail_term_rows + 1)
        self.min_idx = self._align_value_to(self.min_idx, self.scroll_by)

        self.update_max_id()
        self.sel_idx = None

    def tail(self):
        if not self.in_tail_mode:
            return

        self.max_idx = self.data_items_count - 1
        self.min_idx = max(0, self.max_idx - self.avail_term_rows + 1)
        self.sel_idx = None

    def key_enter(self):
        if self.sel_idx is None:
            return None

        self.in_tail_mode = False
        return self.hosts_print_order[self.sel_idx]

    def _align_value_to(self, src_value, aligned_to_value):
        return int(math.ceil(src_value / (aligned_to_value * 1.0)) * aligned_to_value)

def ui_renderer(all_hosts):
    term = gvars['term']

    if term.number_of_colors < 8:
        raise Exception('Your terminal does not support colors')

    with gvars['exit_fullscreen_lock']:
        gvars['exited_fullscreen'] = False
        print(term.enter_fullscreen())

    screen_rows = []
    cmd_err = ['']
    host_data_type = 'parsed'
    sel_hostname = None
    scroller_registry = {
        'parsed': DataScroller(gvars['hosts_print_order'], False),
        'raw': DataScroller(None, True),
    }
    switched_host_data_type = False

    with term.hidden_cursor():
        while not gvars['stop_run']:
            term_size_changed, t_height, t_width = ui_print(term, screen_rows)

            if not term_size_changed:
                time.sleep(0.05)

            screen_rows = []

            _ui_render_header(screen_rows, cmd_err, host_data_type, sel_hostname)

            scroller = scroller_registry[host_data_type]

            if host_data_type == 'parsed':
                scroller.set_data_items_count(len(gvars['hosts_print_order']))
            elif host_data_type == 'raw':
                scroller.set_data_items_count(len(all_hosts[sel_hostname]['raw']))
            else:
                raise NotImplementedError(host_data_type)

            if term_size_changed or not scroller.initialized:
                scroller.set_avail_term_rows(t_height - len(screen_rows)) # skip the header rows
                scroller.reset()

            if switched_host_data_type and host_data_type == 'raw':
                scroller.reset()

            scroller.update_max_id()

            switched_host_data_type = False
            with gvars['keys_pressed_lock']:
                while len(gvars['keys_pressed_list']):
                    key = gvars['keys_pressed_list'].pop(0)
                    cmd_err = ['']

                    min_idx_updated = False
                    if key == '<DOWN>' and host_data_type == 'parsed':
                        scroller.key_down()
                    elif key == '<UP>' and host_data_type == 'parsed':
                        scroller.key_up()
                    elif key == '<PAGEUP>':
                        scroller.key_pageup()
                    elif key == '<PAGEDOWN>':
                        scroller.key_pagedown()
                    elif key == '<HOME>':
                        scroller.key_home()
                    elif key == '<END>':
                        scroller.key_end()
                    elif key == '<ENTER>' and host_data_type == 'parsed':
                        sel_hostname = scroller.key_enter()
                        if sel_hostname is None:
                            cmd_err = [
                                TermCtrl('bold'), TermCtrl('red'),
                                'You need to select a row with the up and down arrows first',
                            ]
                        else:
                            host_data_type = 'raw'

                            switched_host_data_type = True
                            break # mandatory restart
                    elif key == '<ESC>':
                        if host_data_type == 'raw':
                            host_data_type = 'parsed'

                            switched_host_data_type = True
                            break # mandatory restart
                        else:
                            scroller.sel_idx = None
                    elif key == 's' and host_data_type == 'parsed':
                        gvars['stats_show'].rotate(-1)
                    elif key == 'S' and host_data_type == 'parsed':
                        gvars['stats_show'].rotate(1)
                    elif key == 't' and host_data_type == 'parsed':
                        gvars['time_scale'].rotate(-1)
                    elif key == 'T' and host_data_type == 'parsed':
                        gvars['time_scale'].rotate(1)
                    elif key.lower() == 'q':
                        gvars['stop_run'] = True
                    else:
                        if not all(c in string.printable for c in key):
                            # key = '0x' + binascii.hexlify(key.encode('ascii')).decode('ascii')
                            key = ''
                        else:
                            key = f' "{key}"'
                        cmd_err = [TermCtrl('bold'), TermCtrl('red'), f'Unknown key command{key}']

            if switched_host_data_type:
                continue # mandatory restart

            if host_data_type == 'raw':
                scroller.tail()

            if scroller.max_idx < 0:
                # terminal is so small that it doesn't have any space for data, or there
                # is no data
                continue

            _ui_render_all_hosts_data(
                screen_rows, all_hosts, host_data_type,
                scroller.min_idx, scroller.max_idx,
                scroller.sel_idx, sel_hostname, t_width
            )

    # only this thread and Exceptions print any info
    # therefore, it's this thread's task to exit full screen
    exit_fullscreen()

def _exit_fullscreen():
    if not gvars['exited_fullscreen']:
        gvars['exited_fullscreen'] = True
        print(gvars['term'].exit_fullscreen(), end='', flush=True)

def exit_fullscreen():
    if threading.current_thread() != gvars['ui_renderer_thread']: # avoid deadlock
        with gvars['exit_fullscreen_lock']:
            _exit_fullscreen()
    else: # ui_renderer_thread (the only thread which can set "exited_fullscreen = False"
        _exit_fullscreen()

def thread_runner(*args):
    try:
        func = args[0]
        func(*args[1:])
    except:
        gvars['stop_run'] = True

        if threading.current_thread() != gvars['ui_renderer_thread']:
            # this thread must exit very soon, because of the gvars['stop_run'] == False
            # wait for it to call exit_fullscreen(), so that we can successfully
            # print() the exception info
            gvars['ui_renderer_thread'].join(5) # best effort

        exit_fullscreen()
        raise # and print() the exception info

def stdin_processor():
    with curtsies.Input() as input_generator:
        while not gvars['stop_run']:
            e = input_generator.send(0.05)
            if e is None:
                continue # nothing pressed within the timeout period

            with gvars['keys_pressed_lock']:
                key = str(e)
                if key == '<Ctrl-j>':
                    key = '<ENTER>'

                gvars['keys_pressed_list'].append(key)

def sigint_handler(a, b):
    gvars['stop_run'] = True

def populate_hosts():
    ret = {}
    for hostname, cmd in gvars['cmd_args']['ping']:
        if hostname in ret:
            print(f'Error: Duplicate unique name: {hostname}', file=sys.stderr, flush=True)
            sys.exit(1)

        ret[hostname] = {
            'proc': {
                'cmdline': cmd,
                'pid': None,
            },
            'lock': threading.Lock(),
            'stats': {},
            'parsed': [''],
            'raw': [''],
            'raw_complete': False,
            'seen_rx_seq': {},
        }
        
        for k in gvars['stats_show']:
            if not k.endswith('_cnt'):
                ret[hostname]['stats'][k] = None
            else:
                ret[hostname]['stats'][k] = 0

        gvars['hosts_print_order'].append(hostname)

    max_hostname_len = len(max(gvars['hosts_print_order'], key=len))

    if gvars['config']['auto_max_host_id_len']: # fit as much as needed
        gvars['config']['max_host_id_len'] = max_hostname_len
    else:
        if max_hostname_len < gvars['config']['max_host_id_len']:
            # max len is within limits, so shrink limits
            gvars['config']['max_host_id_len'] = max_hostname_len

    return ret

def update_hosts_data():
    workflow = ping_multi_ext.proc.Workflow(gvars['proc_data'], gvars['cmd_args']['timeout'])
    workflow.start_all_processes()

    while not gvars['stop_run']:
        workflow.update_hosts_data(0.05)

def _global_pre_init():
    gvars['stats_show'] = deque([
        'Last', 'Loss%', 'Avg', 'Min', 'Max', 'StDev', 'RX_cnt', 'TX_cnt', 'XX_cnt',
    ])

    gvars['config'] = {
        'auto_max_host_id_len': True,
        'max_host_id_len': 0,
    }

    if gvars['cmd_args']['hosts_max_width'] > 0:
        gvars['config']['auto_max_host_id_len'] = False
        gvars['config']['max_host_id_len'] = gvars['cmd_args']['hosts_max_width']

def _main():
    gvars['term'] = Terminal()

    gvars['stop_run'] = False

    gvars['keys_pressed_list'] = []
    gvars['keys_pressed_lock'] = threading.Lock()

    gvars['exited_fullscreen'] = True # by default we're not in this mode
    gvars['exit_fullscreen_lock'] = threading.Lock()

    gvars['time_scale'] = deque(['success', 'raw', 'numbered'])

    gvars['ui_renderer_thread'] = threading.Thread(name='ui_renderer', target=thread_runner, args=(
        ui_renderer, gvars['proc_data']
    ))

    thr_list = [
        gvars['ui_renderer_thread'],
        threading.Thread(name='stdin_processor', target=thread_runner, args=(stdin_processor,)),
        threading.Thread(name='update_hosts_data', target=thread_runner, args=(update_hosts_data,)),
    ]

    signal.signal(signal.SIGINT, sigint_handler) # CTRL+C

    for thr in thr_list:
        thr.start()

    while not gvars['stop_run']:
        time.sleep(0.05)

    while threading.active_count() > 1:
        for thr in thr_list:
            if thr.is_alive():
                thr.join()
        time.sleep(0.05)

    for host_data in gvars['proc_data'].values():
        if not host_data['proc']['pid']:
            continue
        try:
            os.kill(host_data['proc']['pid'], signal.SIGTERM)
        except ProcessLookupError:
            pass

def main(cmd_args):
    # Execute before we get into fullscreen mode.
    # This way we can easily print() anything and exit then.
    gvars['cmd_args'] = cmd_args
    gvars['hosts_print_order'] = []
    _global_pre_init()
    gvars['proc_data'] = populate_hosts()
    thread_runner(_main)



import ping_multi_ext.lib
import ping_multi_ext.core

def parse_argv():
    parser = ping_multi_ext.lib.argv_parser_base(
        'Ping all hosts from FILE and HOSTs.'
    )

    dval = 1
    parser.add_argument('-W,--wait', dest='wait', metavar='SECS', type=float, default=dval,
        help=f'timeout in seconds to wait for a ping reply; default={dval}')

    dval = 1
    parser.add_argument('-i,--interval', dest='interval', metavar='SECS', type=float, default=dval,
        help=f'time in seconds between sending each request; default={dval}')

    parser.add_argument('-f,--file', dest='file', metavar='FILE',
        help=f'read list of hosts from file')

    dval = 0
    parser.add_argument('--hosts-max-width', type=int, default=dval,
        help=f'maximum width of the hosts column; default={dval}')

    parser.add_argument('host', nargs='*',
        help='host to ping; you can specify this option many times')

    args = vars(parser.parse_args())

    if args['wait'] > args['interval']:
        parser.error('Argument "--wait={:.1f}" cannot be bigger than "--interval={:.1f}"'.format(
            args['wait'], args['interval']
        ))

    hosts = args['host'].copy()

    if args['file'] is not None:
        with open(args['file']) as f:
            for line in f:
                line = line.strip()

                if not len(line) or line.startswith('#'):
                    continue

                hosts.append(line)

    if not len(hosts):
        parser.error('No hosts were specified')

    ping_args = []
    for host in hosts:
        ping_args.append((
            ping_multi_ext.lib.remove_ssh_user(host),
            ping_multi_ext.lib.compose_ping_cmd(host, args),
        ))

    return {
        'timeout': args['wait'],
        'hosts_max_width': args['hosts_max_width'],
        'ping': ping_args,
    }

def main():
    ping_multi_ext.core.main(parse_argv())
